@@ -23,7 +23,7 @@ use super::handle::HandleId;
 use super::handle::{chunk_hash, Chunk};
 use super::packing::Packing;
 use super::state::{ChunkInfo, Pack, PackIndex, RepoState};
-use crate::store::BlockId;
+use crate::store::{BlockId, BlockKey};
 
 /// Encode and decode blocks of data.
 pub trait EncodeBlock {
@@ -80,7 +80,6 @@ pub trait WriteBlock: ReadBlock {
 struct PackingBlockReader<'a> {
     repo_state: &'a RepoState,
     store_state: &'a mut StoreState,
-    pack_size: u32,
 }
 
 impl<'a> ReadBlock for PackingBlockReader<'a> {
@@ -108,7 +107,7 @@ impl<'a> ReadBlock for PackingBlockReader<'a> {
                         .store
                         .lock()
                         .unwrap()
-                        .read_block(pack_index.id)
+                        .read_block(BlockKey::Data(pack_index.id))
                         .map_err(crate::Error::Store)?
                         .ok_or(crate::Error::InvalidData)?;
                     let pack_buffer = self
@@ -151,7 +150,6 @@ impl<'a> ReadBlock for PackingBlockWriter<'a> {
         let mut reader = PackingBlockReader {
             repo_state: self.repo_state,
             store_state: self.store_state,
-            pack_size: self.pack_size,
         };
         reader.read_block(id)
     }
@@ -241,7 +239,7 @@ impl<'a> WriteBlock for PackingBlockWriter<'a> {
                     .store
                     .lock()
                     .unwrap()
-                    .write_block(current_pack.id, encrypted_pack.as_slice())
+                    .write_block(BlockKey::Data(current_pack.id), encrypted_pack.as_slice())
                     .map_err(crate::Error::Store)?;
 
                 // We're starting a new pack, so these need to be reset.
@@ -271,7 +269,7 @@ impl<'a> WriteBlock for PackingBlockWriter<'a> {
                     .store
                     .lock()
                     .unwrap()
-                    .write_block(current_pack.id, encrypted_pack.as_slice())
+                    .write_block(BlockKey::Data(current_pack.id), encrypted_pack.as_slice())
                     .map_err(crate::Error::Store)?;
 
                 // We need to update the pack map in the repository state after all data has been
@@ -296,12 +294,6 @@ struct DirectBlockWriter<'a> {
     state: &'a RepoState,
 }
 
-impl<'a> DirectBlockWriter<'a> {
-    fn new(state: &'a RepoState) -> Self {
-        DirectBlockWriter { state }
-    }
-}
-
 impl<'a> ReadBlock for DirectBlockWriter<'a> {
     fn read_block(&mut self, id: BlockId) -> crate::Result<Vec<u8>> {
         let encoded_block = self
@@ -309,7 +301,7 @@ impl<'a> ReadBlock for DirectBlockWriter<'a> {
             .store
             .lock()
             .unwrap()
-            .read_block(id)
+            .read_block(BlockKey::Data(id))
             .map_err(crate::Error::Store)?
             .ok_or(crate::Error::InvalidData)?;
         self.state.decode_data(encoded_block.as_slice())
@@ -323,7 +315,7 @@ impl<'a> WriteBlock for DirectBlockWriter<'a> {
             .store
             .lock()
             .unwrap()
-            .write_block(id, encoded_block.as_slice())
+            .write_block(BlockKey::Data(id), encoded_block.as_slice())
             .map_err(crate::Error::Store)
     }
 }
@@ -387,10 +379,9 @@ impl<'a> ReadBlock for StoreReader<'a> {
             Packing::None => Box::new(DirectBlockWriter {
                 state: &self.repo_state,
             }),
-            Packing::Fixed(pack_size) => Box::new(PackingBlockReader {
+            Packing::Fixed(_) => Box::new(PackingBlockReader {
                 repo_state: &self.repo_state,
                 store_state: &mut self.store_state,
-                pack_size: *pack_size,
             }),
         };
         read_block.read_block(id)
