@@ -23,17 +23,19 @@ use std::sync::{Arc, RwLock};
 use hex_literal::hex;
 use rmp_serde::{from_read, to_vec};
 use secrecy::ExposeSecret;
+use static_assertions::assert_impl_all;
 use uuid::Uuid;
 
-use crate::store::{BlockId, DataStore};
+use crate::store::{BlockKey, BlockType, DataStore};
 
 use super::chunk_store::{
     EncodeBlock, ReadBlock, ReadChunk, StoreReader, StoreState, StoreWriter, WriteBlock,
 };
 use super::commit::Commit;
 use super::encryption::{Encryption, EncryptionKey, KeySalt, ResourceLimit};
-use super::handle::{chunk_hash, HandleId, HandleIdTable, ObjectHandle, ObjectId};
+use super::handle::{chunk_hash, HandleIdTable, ObjectHandle};
 use super::key::{Key, Keys};
+use super::lock::{unlock_store, Unlock};
 use super::metadata::{Header, RepoInfo, RepoStats};
 use super::object::Object;
 use super::object_store::{ObjectReader, ObjectWriter};
@@ -43,34 +45,6 @@ use super::packing::Packing;
 use super::savepoint::{KeyRestore, RestoreSavepoint, Savepoint};
 use super::state::{InstanceId, InstanceInfo, ObjectState, RepoState};
 
-/// The block ID of the block which stores the repository metadata.
-pub(super) const METADATA_BLOCK_ID: BlockId = BlockId::new(Uuid::from_bytes(hex!(
-    "8691d360 29c6 11ea 8bc1 2fc8cfe66f33"
-)));
-
-/// The block ID of the block which stores the repository format version.
-pub(super) const VERSION_BLOCK_ID: BlockId = BlockId::new(Uuid::from_bytes(hex!(
-    "cbf28b1c 3550 11ea 8cb0 87d7a14efe10"
-)));
-
-/// Return a list of blocks in the data store excluding those used to store metadata.
-fn list_data_blocks(state: &RepoState) -> crate::Result<Vec<BlockId>> {
-    let all_blocks = state
-        .store
-        .lock()
-        .unwrap()
-        .list_blocks()
-        .map_err(crate::Error::Store)?;
-
-    Ok(all_blocks
-        .iter()
-        .copied()
-        .filter(|id| {
-            *id != METADATA_BLOCK_ID && *id != VERSION_BLOCK_ID && *id != state.metadata.header_id
-        })
-        .collect())
-}
-
 /// An object store which maps keys to seekable binary blobs.
 ///
 /// See [`crate::repo::key`] for more information.
@@ -101,6 +75,8 @@ pub struct KeyRepo<K: Key> {
     pub(super) transaction_id: Arc<Uuid>,
 }
 
+assert_impl_all!(KeyRepo<()>: Send, Sync);
+
 impl<K: Key> OpenRepo for KeyRepo<K> {
     type Key = K;
 
@@ -128,13 +104,6 @@ impl<K: Key> OpenRepo for KeyRepo<K> {
 }
 
 impl<K: Key> KeyRepo<K> {
-    /// Return the `object_id` for the object with the given `handle_id`.
-    fn object_id(&self, handle_id: HandleId) -> ObjectId {
-        let state = self.state.read().unwrap();
-        let repo_id = state.metadata.id;
-        ObjectId::new(repo_id, self.instance_id, handle_id)
-    }
-
     /// Return whether there is an object with the given `key` in this repository.
     pub fn contains<Q>(&self, key: &Q) -> bool
     where
@@ -150,7 +119,6 @@ impl<K: Key> KeyRepo<K> {
     pub fn insert(&mut self, key: K) -> Object {
         self.remove(&key);
         let handle_id = self.handle_table.next();
-        let object_id = self.object_id(handle_id);
         let handle = ObjectHandle {
             id: handle_id,
             extents: Vec::new(),
@@ -160,7 +128,7 @@ impl<K: Key> KeyRepo<K> {
             .objects
             .entry(key)
             .or_insert_with(|| Arc::new(RwLock::new(handle)));
-        Object::new(&self.state, handle, object_id)
+        Object::new(&self.state, handle)
     }
 
     /// Remove the given object `handle` from the repository.
@@ -210,8 +178,7 @@ impl<K: Key> KeyRepo<K> {
         Q: Eq + Hash + ?Sized,
     {
         let handle = self.objects.get(key)?;
-        let handle_id = handle.read().unwrap().id;
-        Some(Object::new(&self.state, handle, self.object_id(handle_id)))
+        Some(Object::new(&self.state, handle))
     }
 
     /// Return an iterator over all the keys of objects in this repository.
@@ -376,7 +343,7 @@ impl<K: Key> KeyRepo<K> {
             .store
             .lock()
             .unwrap()
-            .write_block(header_id, encoded_header.as_slice())
+            .write_block(BlockKey::Header(header_id), encoded_header.as_slice())
             .map_err(crate::Error::Store)?;
         state.metadata.header_id = header_id;
 
@@ -387,7 +354,7 @@ impl<K: Key> KeyRepo<K> {
             .store
             .lock()
             .unwrap()
-            .write_block(METADATA_BLOCK_ID, &serialized_metadata)
+            .write_block(BlockKey::Super, &serialized_metadata)
             .map_err(crate::Error::Store)?;
         Ok(())
     }
@@ -735,7 +702,7 @@ impl<K: Key> Commit for KeyRepo<K> {
             .store
             .lock()
             .unwrap()
-            .read_block(state.metadata.header_id)
+            .read_block(BlockKey::Header(state.metadata.header_id))
             .map_err(crate::Error::Store)?
             .ok_or(crate::Error::Corrupt)?;
         let serialized_header = state.decode_data(encoded_header.as_slice())?;
@@ -755,7 +722,7 @@ impl<K: Key> Commit for KeyRepo<K> {
             .store
             .lock()
             .unwrap()
-            .read_block(state.metadata.header_id)
+            .read_block(BlockKey::Header(state.metadata.header_id))
             .map_err(crate::Error::Store)?
             .ok_or(crate::Error::Corrupt)?;
         let serialized_header = state.decode_data(encoded_header.as_slice())?;
@@ -778,16 +745,27 @@ impl<K: Key> Commit for KeyRepo<K> {
         // Remove all blocks from the data store which are unreferenced.
         match &state.metadata.config.packing {
             Packing::None => {
-                // When packing is disabled, we can just remove the unreferenced blocks from the
-                // data store directly.
-                let block_ids = list_data_blocks(&state)?;
+                // When packing is disabled, we can just remove the unreferenced data blocks from
+                // the data store directly.
+                {
+                    let block_ids = state
+                        .store
+                        .lock()
+                        .unwrap()
+                        .list_blocks(BlockType::Data)
+                        .map_err(crate::Error::Store)?;
 
-                let mut store = state.store.lock().unwrap();
-                for block_id in block_ids {
-                    if !referenced_blocks.contains(&block_id) {
-                        store.remove_block(block_id).map_err(crate::Error::Store)?;
+                    let mut store = state.store.lock().unwrap();
+                    for block_id in block_ids {
+                        if !referenced_blocks.contains(&block_id) {
+                            store
+                                .remove_block(BlockKey::Data(block_id))
+                                .map_err(crate::Error::Store)?;
+                        }
                     }
                 }
+
+                drop(state);
             }
             Packing::Fixed(_) => {
                 // When packing is enabled, we need to repack the packs which contain unreferenced
@@ -815,7 +793,13 @@ impl<K: Key> Commit for KeyRepo<K> {
                 let mut blocks_to_repack = Vec::new();
 
                 // Iterate over the IDs of packs which are contained in the data store.
-                for pack_id in list_data_blocks(&state)? {
+                let data_blocks = state
+                    .store
+                    .lock()
+                    .unwrap()
+                    .list_blocks(BlockType::Data)
+                    .map_err(crate::Error::Store)?;
+                for pack_id in data_blocks {
                     match packs_to_blocks.get(&pack_id) {
                         Some(contained_blocks) => {
                             let contains_unreferenced_blocks = contained_blocks
@@ -850,7 +834,9 @@ impl<K: Key> Commit for KeyRepo<K> {
                 {
                     let mut store = state.store.lock().unwrap();
                     for pack_id in packs_to_remove {
-                        store.remove_block(pack_id).map_err(crate::Error::Store)?;
+                        store
+                            .remove_block(BlockKey::Data(pack_id))
+                            .map_err(crate::Error::Store)?;
                     }
                 }
 
@@ -890,6 +876,66 @@ impl<K: Key> Commit for KeyRepo<K> {
             }
         }
 
+        // Remove old unreferenced headers from the data store.
+        {
+            let state = self.state.read().unwrap();
+            let mut store = state.store.lock().unwrap();
+            let unreferenced_headers = store
+                .list_blocks(BlockType::Header)
+                .map_err(crate::Error::Store)?
+                .into_iter()
+                .filter(|&block_id| block_id != state.metadata.header_id);
+            for block_id in unreferenced_headers {
+                store
+                    .remove_block(BlockKey::Header(block_id))
+                    .map_err(crate::Error::Store)?;
+            }
+        }
+
         Ok(())
     }
 }
+
+impl<K: Key> Unlock for KeyRepo<K> {
+    fn unlock(&self) -> crate::Result<()> {
+        let state = self.state.read().unwrap();
+        let mut store = state.store.lock().unwrap();
+        unlock_store(&mut *store, state.lock_id)
+    }
+
+    fn is_locked(&self) -> crate::Result<bool> {
+        let state = self.state.read().unwrap();
+        let mut store = state.store.lock().unwrap();
+        store
+            .read_block(BlockKey::Lock(state.lock_id))
+            .map_err(crate::Error::Store)
+            .map(|result| result.is_some())
+    }
+
+    fn context(&self) -> crate::Result<Vec<u8>> {
+        let state = self.state.read().unwrap();
+        let mut store = state.store.lock().unwrap();
+        let encrypted_context = store
+            .read_block(BlockKey::Lock(state.lock_id))
+            .map_err(crate::Error::Store)?
+            .ok_or(crate::Error::NotLocked)?;
+        state
+            .metadata
+            .config
+            .encryption
+            .decrypt(&encrypted_context, &state.master_key)
+    }
+
+    fn update_context(&self, context: &[u8]) -> crate::Result<()> {
+        let state = self.state.read().unwrap();
+        let mut store = state.store.lock().unwrap();
+        let encrypted_context = state
+            .metadata
+            .config
+            .encryption
+            .encrypt(context, &state.master_key);
+        store
+            .write_block(BlockKey::Lock(state.lock_id), &encrypted_context)
+            .map_err(crate::Error::Store)
+    }
+}
